# -*-coding:utf-8 -*-
u"""
:创建时间: 2022/4/25 19:38
:作者: 苍之幻灵
:我的主页: https://cpcgskill.com
:QQ: 2921251087
:爱发电: https://afdian.net/@Phantom_of_the_Cang
:aboutcg: https://www.aboutcg.org/teacher/54335
:bilibili: https://space.bilibili.com/351598127

"""
from __future__ import unicode_literals, print_function

Alpha, Beta, Rc, R = (0, 100, 200, 300)


def adv_ver_conv(i):
    adv_ver_conv_table = {
        Alpha: 'alpha',
        Beta: 'beta',
        Rc: 'rc',
        R: "r",
    }
    return adv_ver_conv_table[i]


class VersionInfo(object):
    def __init__(self, major, minor, patch, adv_ver=R, adv_ver_index=0):
        self.major = major
        self.minor = minor
        self.patch = patch
        self.adv_ver = adv_ver
        self.adv_ver_index = adv_ver_index

    def to_str(self):
        s = "{}.{}.{}".format(self.major, self.minor, self.patch)
        if self.adv_ver < R:
            s += "-{}.{}".format(adv_ver_conv(self.adv_ver), self.adv_ver_index)
        return s


def ver_info():
    return VersionInfo(major=3, minor=2, patch=1, adv_ver=R, adv_ver_index=0)


def ver_str():
    return ver_info().to_str()


if __name__ == "__main__":
    print(ver_str())



# -*-coding:utf-8 -*-
u"""
:创建时间: 2022/3/18 19:04
:作者: 苍之幻灵
:我的主页: https://cpcgskill.com
:QQ: 2921251087
:爱发电: https://afdian.net/@Phantom_of_the_Cang
:aboutcg: https://www.aboutcg.org/teacher/54335
:bilibili: https://space.bilibili.com/351598127

"""
from __future__ import unicode_literals, print_function
import array

from cpapi.all import *
import cpapi.all as om
import cpapi.utils as omtl
import maya.api.OpenMaya as om2
import maya.api.OpenMayaAnim as om2aim
from cpmel._object_types.core import Node, Transform, Shape, other_node_cls, new_object, Attr, Component
from cpmel._args_conv import arg_conv
from cpmel.exc import *


@other_node_cls("kJoint")
class Joint(Transform):
    pass


@other_node_cls("kMesh")
class Mesh(Shape):
    component_configs = [
        ('vtx', 'vtx[*]'),
        ('e', 'e[*]'),
        ('f', 'f[*]'),
        ('vtxFace', 'vtxFace[*][*]'),
        ('map', 'map[*]'),
    ]
    __slots__ = tuple(['ref'] + [i[0] for i in component_configs])

    def api1_m_fn(self):
        return MFnMesh(self.api1_m_dag_path())

    def api2_m_fn(self):
        return om2.MFnMesh(self.api2_m_dag_path())

    def set_point(self, i, p, ws=True):
        i = arg_conv(i)
        p = arg_conv(p)
        ws = arg_conv(ws)
        new_object("{}.vtx[{}]".format(self.name(), i)).set_point(p, ws)
        return self

    def get_point(self, i, ws=True):
        i = arg_conv(i)
        ws = arg_conv(ws)
        return new_object("{}.vtx[{}]".format(self.name(), i)).get_point(ws)

    def set_points(self, points, ws=True):
        """
        设置所有顶点位置

        :param points: 点位置列表
        :param ws: 是否为世界空间
        :return:
        """
        [i.set_point(p, ws) for i, p in zip(new_object("{}.vtx[*]".format(self.name())), points)]
        return self

    def get_points(self, ws=True):
        return [i.get_point(ws) for i in new_object("{}.vtx[*]".format(self.name()))]


@other_node_cls('kNurbsCurve')
class Curve(Shape):
    component_configs = [
        ('cv', 'cv[*]'),
        ('ep', 'ep[*]'),
        ('u', 'u[*]'),
    ]
    __slots__ = tuple(['ref'] + [i[0] for i in component_configs])

    def api1_m_fn(self):
        return om.MFnNurbsCurve(self.api1_m_dag_path())

    def api2_m_fn(self):
        return om2.MFnNurbsCurve(self.api2_m_dag_path())

    def set_point(self, i, p, ws=True):
        i = arg_conv(i)
        p = arg_conv(p)
        ws = arg_conv(ws)
        new_object("{}.cv[{}]".format(self.name(), i)).set_point(p, ws)
        return self

    def get_point(self, i, ws=True):
        i = arg_conv(i)
        ws = arg_conv(ws)
        return new_object("{}.cv[{}]".format(self.name(), i)).get_point(ws)

    def set_points(self, points, ws=True):
        """
        设置所有顶点位置

        :param points: 点位置列表
        :param ws: 是否为世界空间
        :return:
        """
        [i.set_point(p, ws) for i, p in zip(new_object("{}.cv[*]".format(self.name())), points)]
        return self

    def get_points(self, ws=True):
        return [i.get_point(ws) for i in new_object("{}.cv[*]".format(self.name()))]

    def num_cvs(self):
        u"""
        返回此曲线的CV数

        :return:
        """
        return self.api1_m_fn().numCVs()

    def num_spans(self):
        u"""
        返回此曲线的跨度数

        :return:
        """
        return self.api1_m_fn().numSpans()

    def num_knots(self):
        u"""
        返回此曲线的结数

        :return:
        """
        return self.api1_m_fn().numKnots()

    def degree(self):
        u"""
        返回此曲线的度数

        :return:
        """
        return self.api1_m_fn().degree()


@other_node_cls('kNurbsSurface')
class Surface(Shape):
    component_configs = [
        ('cv', 'cv[*]'),
        ('u', 'u[*]'),
        ('v', 'v[*]'),
        ('uv', 'uv[*][*]'),
        ('sf', 'sf[*][*]'),
    ]
    __slots__ = tuple(['ref'] + [i[0] for i in component_configs])

    def api1_m_fn(self):
        return om.MFnNurbsSurface(self.api1_m_dag_path())

    def api2_m_fn(self):
        return om2.MFnNurbsSurface(self.api2_m_dag_path())

    def set_point(self, ia, ib, p, ws=True):
        ia = arg_conv(ia)
        ib = arg_conv(ib)
        p = arg_conv(p)
        ws = arg_conv(ws)
        new_object("{}.cv[{}][{}]".format(self.name(), ia, ib)).set_point(p, ws)
        return self

    def get_point(self, ia, ib, ws=True):
        ia = arg_conv(ia)
        ib = arg_conv(ib)
        ws = arg_conv(ws)
        return new_object("{}.cv[{}][{}]".format(self.name(), ia, ib)).get_point(ws)

    def set_points(self, points, ws=True):
        """
        设置所有顶点位置

        :param points: 点位置列表
        :param ws: 是否为世界空间
        :return:
        """
        [i.set_point(p, ws) for i, p in zip(new_object("{}.cv[*]".format(self.name())), points)]
        return self

    def get_points(self, ws=True):
        return [i.get_point(ws) for i in new_object("{}.cv[*]".format(self.name()))]


@other_node_cls("kSkinClusterFilter")
class SkinClusterFilter(Node):
    def api1_m_fn(self):
        return om.MFnSkinCluster(self.api1_node_object())

    def api2_m_fn(self):
        return om2aim.MFnSkinCluster(self.api2_node_object())

    def get_weights(self, geo, inf):
        """
        获得骨骼蒙皮权重

        :param geo: 组件 例:xxx.vtx
        :param inf: 关节索引
        :return:
        :rtype:list
        """

        geo = new_object(arg_conv(geo))
        if not isinstance(geo, Component):
            raise CPMelException("需要组件")
        p, c = geo.api1_m_component()

        try:
            inf = array.array('i', inf)
        except TypeError:
            raise CPMelException(u"inf 不是整形数组")

        m_inf = MIntArray()
        ScriptUtil = om.MScriptUtil()
        ScriptUtil.createIntArrayFromList(inf, m_inf)
        m_weigth = MDoubleArray()
        self.api1_m_fn().getWeights(p, c, m_inf, m_weigth)
        return array.array('f', m_weigth)

    def unsafe_set_weights(self, geo, inf, weights):
        """
        获得骨骼蒙皮权重

        :param geo: 组件 例:xxx.vtx
        :param inf: 关节索引
        :param weights: 权重数组
        :return:
        """
        geo = new_object(arg_conv(geo))
        if not isinstance(geo, Component):
            raise CPMelException("需要组件")
        p, c = geo.api1_m_component()

        try:
            inf = array.array('i', inf)
        except TypeError:
            raise CPMelException("inf 不是整形数组")

        try:
            weights = array.array('f', weights)
        except TypeError:
            raise CPMelException("weight 不是浮点数组")

        m_inf = MIntArray()
        ScriptUtil = om.MScriptUtil()
        ScriptUtil.createIntArrayFromList(inf, m_inf)
        m_weights = MDoubleArray()
        for i in weights:
            m_weights.append(i)
        # 撤销的权重
        un_weights = MDoubleArray()

        self.api1_m_fn().setWeights(p, c, m_inf, m_weights, True, un_weights)
        return self

    def get_blend_weights(self, geo):
        """
        获得权重已混合权重

        :param geo: 组件
        :return:
        :rtype:list
        """
        geo = new_object(arg_conv(geo))
        if not isinstance(geo, Component):
            raise CPMelException("需要组件")
        p, c = geo.api1_m_component()

        weights = MDoubleArray()
        self.api1_m_fn().getBlendWeights(p, c, weights)
        return array.array('f', weights)

    def unsafe_set_blend_weights(self, geo, weights):
        """
        设置权重已混合权重

        :param geo: 组件
        :param weights: 权重数组
        :return:
        """
        geo = new_object(arg_conv(geo))
        if not isinstance(geo, Component):
            raise CPMelException("需要组件")
        p, c = geo.api1_m_component()

        try:
            weights = array.array('f', weights)
        except TypeError:
            raise CPMelException("weight 不是浮点数组")

        m_weights = MDoubleArray()
        for i in weights:
            m_weights.append(i)

        self.api1_m_fn().setBlendWeights(p, c, m_weights)

        return self



# -*-coding:utf-8 -*-
u"""
:创建时间: 2022/3/16 18:06
:作者: 苍之幻灵
:我的主页: https://cpcgskill.com
:QQ: 2921251087
:爱发电: https://afdian.net/@Phantom_of_the_Cang
:aboutcg: https://www.aboutcg.org/teacher/54335
:bilibili: https://space.bilibili.com/351598127

"""
from __future__ import unicode_literals, print_function
import sys
from collections import Iterable, Callable

from cpapi.all import (MUuid,
                       MVector, MPoint, MMatrix,
                       MFloatVector, MFloatPoint, MFloatMatrix,
                       MEulerRotation,
                       MDagPath)
import maya.OpenMaya as om1

from cpmel.exc import *


def m_uuid_to_melobject(i):
    return i.asString()


def m_vector_to_melobject(v):
    return v.x, v.y, v.z


def m_point_to_melobject(p):
    if p.w == 1:
        return p.x, p.y, p.z
    else:
        return p.x, p.y, p.z, p.w


def m_matrix_to_melobject(m):
    return tuple((m(r, c) for r in range(4) for c in range(4)))


def m_euler_rotation_to_melobject(e):
    return e.x, e.y, e.z


def m_dag_path_to_melobject(p):
    return p.fullPathName()


def object_type_to_melobject(n):
    return n.__melobject__()


def ui_to_melobject(ui):
    return ui.__melobject__()


# 类型编码表
type_conv_table = {
    om1.MUuid: m_uuid_to_melobject,
    MUuid: m_uuid_to_melobject,

    om1.MPoint: m_point_to_melobject,
    om1.MFloatPoint: m_point_to_melobject,
    MPoint: m_point_to_melobject,
    MFloatPoint: m_point_to_melobject,

    om1.MVector: m_vector_to_melobject,
    om1.MFloatVector: m_vector_to_melobject,
    MVector: m_vector_to_melobject,
    MFloatVector: m_vector_to_melobject,

    om1.MMatrix: m_matrix_to_melobject,
    om1.MFloatMatrix: m_matrix_to_melobject,
    MMatrix: m_matrix_to_melobject,
    MFloatMatrix: m_matrix_to_melobject,

    om1.MEulerRotation: m_euler_rotation_to_melobject,
    MEulerRotation: m_euler_rotation_to_melobject,

    om1.MDagPath: m_dag_path_to_melobject,
    MDagPath: m_dag_path_to_melobject,

    # BaseType: object_type_to_melobject,
    # UI: ui_to_melobject,
}

# 白名单
whitelist = {
    bytes,
    str,
    int,
    float,
    bool,
    type(None),
}
if sys.version_info.major == 2:
    whitelist.add(unicode)
    whitelist.add(long)


def arg_conv(arg):
    arg_t = type(arg)
    if arg_t in whitelist:
        return arg
    # if issubclass(arg_t, BaseType):
    #     return object_type_to_melobject(arg)
    conv_func = type_conv_table.get(arg_t)
    if conv_func is not None:
        return conv_func(arg)
    if hasattr(arg, '__melobject__'):
        return arg.__melobject__()
    if isinstance(arg, Iterable):
        return [arg_conv(i) for i in arg]
    if isinstance(arg, Callable):
        return arg
    raise ArgConvTypeException("args {}: {}".format(type(arg), arg))


def kwarg_conv(arg):
    arg_t = type(arg)
    if arg_t in whitelist:
        return arg
    # if issubclass(arg_t, BaseType):
    #     return object_type_to_melobject(arg)
    conv_func = type_conv_table.get(arg_t)
    if conv_func is not None:
        return conv_func(arg)
    if hasattr(arg, '__melobject__'):
        return arg.__melobject__()
    if isinstance(arg, Iterable):
        return [kwarg_conv(i) for i in arg]
    if isinstance(arg, Callable):
        return arg
    raise ArgConvTypeException("kwargs {}: {}".format(type(arg), arg))


def args_conv(args):
    return [arg_conv(i) for i in args]


def kwargs_conv(kwargs):
    return {k: kwarg_conv(arg) for k, arg in kwargs.items()}


_str_t = type('')


def ret_conv(ret):
    if isinstance(ret, _str_t):
        try:
            return new_object(ret)
        except CPMelException as ex:
            return ret

    # if isinstance(ret, tuple) and len(ret) == 3 and tuple((type(i) for i in ret)) == (float, float, float):
    if isinstance(ret, tuple) and len(ret) == 3:
        return MVector(ret[0], ret[1], ret[2])
    if isinstance(ret, list):
        return [ret_conv(i) for i in ret]
    return ret


def ui_ret_conv(ret):
    if isinstance(ret, _str_t):
        return UI(ret)
    return ret


from cpmel._object_types.core import new_object, UI

if __name__ == "__main__":
    print(list(
        args_conv([1, 2, MFloatPoint(1, 1, 1), MMatrix.newMatrix([
            [0, 0, 0, 1],
            [1, 0, 0, 1],
            [0, 1, 0, 1],
            [0, 0, 1, 1],
        ])])
    ))

    print(
        kwargs_conv({
            "n": "wing",
            "tv": [MVector(1, 1, 1), MVector(1, 1, 1)],
            "rgb": MFloatPoint(1, 1, 1)
        })
    )



# -*-coding:utf-8 -*-
u"""
:创建时间: 2022/3/15 4:49
:作者: 苍之幻灵
:我的主页: https://cpcgskill.com
:QQ: 2921251087
:爱发电: https://afdian.net/@Phantom_of_the_Cang
:aboutcg: https://www.aboutcg.org/teacher/54335
:bilibili: https://space.bilibili.com/351598127

"""



# -*-coding:utf-8 -*-
u"""
:创建时间: 2022/3/17 4:11
:作者: 苍之幻灵
:我的主页: https://cpcgskill.com
:QQ: 2921251087
:爱发电: https://afdian.net/@Phantom_of_the_Cang
:aboutcg: https://www.aboutcg.org/teacher/54335
:bilibili: https://space.bilibili.com/351598127

"""
from __future__ import unicode_literals, print_function
import sys

import maya.cmds as mc

try:
    from PySide2.QtWidgets import QWidget
    from shiboken2 import wrapInstance
except ImportError:
    from PySide.QtGui import QWidget
    from shiboken import wrapInstance

import cpapi.utils as omtl
import cpapi.OpenMaya as om
import cpapi.OpenMayaUI as omui
import maya.api.OpenMaya as om2

import cpref.object_ref as cr
from cpref.object_ref import Ref
from cpref import make_ref

from cpmel.exc import *

_ptr = long if sys.version_info.major == 2 else int
_any_int = (int, long) if sys.version_info.major == 2 else int


class BaseType(object):
    __slots__ = ('ref',)

    def __init__(self, ref):
        self.ref = ref  # type: Ref

    def __melobject__(self):
        return self.ref.as_string()

    def __melobject_list__(self):
        return self.ref.as_string_list()

    def assert_valid(self):
        if self.ref.is_null():
            raise RefException("引用已丢失")
        return self

    def name(self):
        self.assert_valid()
        return self.ref.unsafe_as_string_list()[0]


class Node(BaseType):
    __slots__ = ('ref',)

    def type(self):
        return self.api2_m_fn_dependency_node().typeName

    def __hash__(self):
        return om2.MObjectHandle(self.api2_node_object()).hashCode()

    def __eq__(self, other):
        if isinstance(other, BaseType):
            return hash(self) == hash(other)
        return False

    def __ne__(self, other):
        return not self == other

    def __str__(self):
        return self.name()

    def __repr__(self):
        return "{}({})".format(mc.objectType(self.name()), repr(self.name()))

    def attr(self, name):
        return new_object("{}.{}".format(self.name(), name))

    def __getitem__(self, item):
        return self.attr(item).get_value()

    def __setitem__(self, item, value):
        return self.attr(item).set_value(value)

    def __getattr__(self, item):
        try:
            return self.attr(item)
        except CPMelException:
            raise AttributeError("{} object has no attribute '{}'".format(self.__class__.__name__, item))

    def api1_node_object(self):
        self.assert_valid()

        sel = om.MSelectionList()
        sel.add(self.ref.as_string())
        obj = om.MObject()
        sel.getDependNode(0, obj)
        return obj

    def api2_node_object(self):
        self.assert_valid()

        return self.ref.unsafe_m_selection_list().getDependNode(0)

    def api1_m_fn_dependency_node(self):
        return om.MFnDependencyNode(self.api1_node_object())

    def api2_m_fn_dependency_node(self):
        return om2.MFnDependencyNode(self.api2_node_object())

    def api1_m_fn(self):
        return self.api1_m_fn_dependency_node()

    def api2_m_fn(self):
        return self.api2_m_fn_dependency_node()

    def rename(self, name):
        self.api2_m_fn_dependency_node().setName(name)
        return self

    def node_name(self):
        return self.api2_m_fn_dependency_node().name()

    def copy(self):
        u"""

        :return:
        :rtype: Node
        """
        return new_object(mc.duplicate(str(self), rc=True)[0])


class DagNode(Node):
    __slots__ = ('ref',)

    def api1_m_dag_path(self):
        self.assert_valid()

        sel = om.MSelectionList()
        sel.add(self.ref.as_string())
        p = om.MDagPath()
        sel.getDagPath(0, p)
        return p

    def api2_m_dag_path(self):
        self.assert_valid()

        return self.ref.unsafe_m_dag_path()

    def get_translation(self, ws=True):
        ws = arg_conv(ws)
        return om.MVector(*mc.xform(self.name(), q=True, t=True, ws=ws))

    def set_translation(self, p, ws=True):
        p = arg_conv(p)
        ws = arg_conv(ws)
        mc.xform(self.name(), t=(p[0], p[1], p[2]), ws=ws)

    translation = property(get_translation, set_translation)

    def get_rotation(self, ws=True):
        ws = arg_conv(ws)
        return om.MEulerRotation(*mc.xform(self.name(), q=True, ro=True, ws=ws))

    def set_rotation(self, r, ws=True):
        r = arg_conv(r)
        ws = arg_conv(ws)
        mc.xform(self.name(), ro=(r[0], r[1], r[2]), ws=ws)

    rotation = property(get_rotation, set_rotation)

    def get_scale(self, ws=True):
        ws = arg_conv(ws)
        return mc.xform(self.name(), q=True, s=True, ws=ws)

    def set_scale(self, s, ws=True):
        s = arg_conv(s)
        ws = arg_conv(ws)
        mc.xform(self.name(), s=(s[0], s[1], s[2]), ws=ws)

    scale = property(get_scale, set_scale)

    def get_matrix(self, ws=True):
        ws = arg_conv(ws)
        return omtl.new_matrix(mc.xform(self.name(), q=True, m=True, ws=ws))

    def set_matrix(self, m, ws=True):
        m = arg_conv(m)
        ws = arg_conv(ws)
        mc.xform(self.name(), m=m, ws=ws)
        return self

    matrix = property(get_matrix, set_matrix)

    def api1_m_fn(self):
        return om.MFnDagNode(self.api1_m_dag_path())

    def api2_m_fn(self):
        return om2.MFnDagNode(self.api2_m_dag_path())

    def full_path_name(self):
        self.assert_valid()
        return self.ref.unsafe_full_path_name()

    def set_parent(self, p):
        self_ = arg_conv(self)
        p = arg_conv(p)
        mc.parent(self_, p)
        return self

    def add_child(self, *c):
        self_ = arg_conv(self)
        for i in c:
            i = arg_conv(i)
            mc.parent(i, self_)
        return self

    def get_parent(self):
        u"""
        获得父对象

        :return:
        :rtype: DagNode
        """
        p = mc.listRelatives(self.name(), p=True, pa=True)
        if p is None:
            return p
        return new_object(p[0])

    @property
    def parent(self):
        return self.get_parent()

    @parent.setter
    def parent(self, v):
        self.set_parent(v)

    def show(self):
        self['visibility'] = True
        return self

    def hide(self):
        self['visibility'] = False
        return self


class Transform(DagNode):
    def __init__(self, ref):
        super(Transform, self).__init__(ref)
        p = self.api2_m_dag_path()  # type: om2.MDagPath
        self.shapes = [new_object(p.extendToShape(i).fullPathName()) for i in range(p.numberOfShapesDirectlyBelow())]
        if len(self.shapes):
            self.shape = self.shapes[0]
        else:
            self.shape = None
        if self.shape is not None:
            for n, _ in self.shape.component_configs:
                setattr(self, n, getattr(self.shape, n))


class Shape(DagNode):
    component_configs = []

    def __init__(self, ref):
        super(Shape, self).__init__(ref)
        for c_n, init_str in type(self).component_configs:
            setattr(self, c_n, new_object("{}.{}".format(self.name(), init_str)))


class Attr(BaseType):
    def __hash__(self):
        return om2.MObjectHandle(self.api2_m_plug().attribute()).hashCode()

    def __eq__(self, other):
        if isinstance(other, BaseType):
            return hash(self) == hash(other)
        return False

    def __ne__(self, other):
        return not self == other

    def __lshift__(self, other):
        a = arg_conv(self)
        b = arg_conv(other)
        mc.connectAttr(b, a, f=True)
        return new_object(b)

    def __rshift__(self, other):
        a = arg_conv(self)
        b = arg_conv(other)
        mc.connectAttr(a, b, f=True)
        return new_object(b)

    def attr(self, name):
        return new_object("{}.{}".format(self.name(), name))

    def __getattr__(self, item):
        try:
            return self.attr(item)
        except CPMelException:
            raise AttributeError("{} object has no attribute '{}'".format(self.__class__.__name__, item))

    def __getitem__(self, item):
        try:
            if isinstance(item, _any_int):
                return new_object("{}[{}]".format(self.name(), item))
            else:
                raise TypeError('attribute indices must be integers, not {}'.format(item.__class__.__name__))
        except CPMelException:
            raise AttributeError("{} object has no item '{}'".format(self.__class__.__name__, item))

    def api1_m_plug(self):
        self.assert_valid()

        sel = om.MSelectionList()
        sel.add(self.ref.unsafe_as_string_list()[0])
        p = om.MPlug()
        sel.getPlug(0, p)
        return p

    def api2_m_plug(self):
        self.assert_valid()
        return self.ref.unsafe_m_selection_list().getPlug(0)

    def node(self):
        return new_object(self.name().split('.', 1)[0])

    def type(self):
        try:
            return mc.getAttr(self.name(), type=True)
        except RuntimeError:
            return None

    def lock(self):
        mc.setAttr(self.name(), lock=True)
        return self

    def unlock(self):
        mc.setAttr(self.name(), lock=False)
        return self

    def is_lock(self):
        return mc.getAttr(self.name(), lock=False)

    def connect(self, b):
        self_ = arg_conv(self)
        b = arg_conv(b)
        mc.connectAttr(self_, b, f=True)
        return self

    def disconnect(self, b):
        self_ = arg_conv(self)
        b = arg_conv(b)
        mc.disconnectAttr(self_, b)
        return self

    def get_value(self):
        t = self.type()
        v = mc.getAttr(self.name())
        if t in {"double3", "float3"}:
            return om.MVector(*(v[0]))
        elif t in {'reflectanceRGB', 'short3', 'spectrumRGB', 'long3'}:
            return v[0]
        elif t == 'matrix':
            return omtl.new_matrix((
                v[0:4],
                v[4:8],
                v[8:12],
                v[12:16]
            ))
        return v

    def set_value(self, val):
        val = arg_conv(val)
        t = self.type()
        if t in {'float3', 'short3', 'spectrumRGB', 'reflectanceRGB', 'double3', 'long3'}:
            mc.setAttr(self.name(), *val, type=t)
        elif t in {'string', 'Int32Array', 'doubleArray'}:
            mc.setAttr(self.name(), val, type=t)
        elif t in {'vectorArray', 'pointArray', 'stringArray', 'componentList'}:
            mc.setAttr(self.name(), len(val), *val)
        elif t == "matrix":
            mc.setAttr(self.name(), *val, type=t)
        else:
            mc.setAttr(self.name(), val)
        return self

    def __str__(self):
        return self.name()

    def __repr__(self):
        return "{}('{}', attr='{}')".format(mc.objectType(self.node().name()), *self.name().split('.', 1))


class Component(BaseType):
    def __hash__(self):
        return om2.MObjectHandle(self.api2_m_component()[1]).hashCode()

    def __eq__(self, other):
        if isinstance(other, BaseType):
            return hash(self) == hash(other)
        return False

    def __ne__(self, other):
        return not self == other

    def api1_m_component(self):
        self.assert_valid()

        sel = om.MSelectionList()
        sel.add(self.ref.unsafe_as_string_list()[0])
        p = om.MDagPath()
        c = om.MObject()
        sel.getDagPath(0, p, c)
        return p, c

    def api2_m_component(self):
        self.assert_valid()
        return self.ref.unsafe_m_selection_list().getComponent(0)

    def node(self):
        return new_object(self.name().split('.', 1)[0])

    def get_point(self, ws=True):
        return om.MVector(*mc.xform(self.name(), q=True, t=True, ws=ws))

    def set_point(self, p, ws=True):
        p = arg_conv(p)
        mc.xform(self.name(), t=(p[0], p[1], p[2]), ws=ws)
        return self

    point = property(get_point, set_point)

    def __str__(self):
        return self.name()

    def __repr__(self):
        obj = self.__melobject_list__()
        return "{}('{}', component={})".format(
            mc.objectType(self.node().name()),
            self.name().split('.', 1)[0],
            repr(list(obj) if len(obj) > 1 else obj[0])
        )

    def __iter__(self):
        return (new_object(i) for i in mc.ls(self.__melobject__(), fl=True))


class UI(object):
    __slots__ = ('ui',)

    def __init__(self, ui):
        self.ui = ui

    def __hash__(self):
        return hash(self.ui)

    def __eq__(self, other):
        if isinstance(other, UI):
            return hash(self) == hash(other)
        return False

    def __ne__(self, other):
        return not self == other

    def get_widget(self):
        u"""
        :return:
        """
        return wrapInstance(_ptr(omui.MQtUtil.findWindow(self.ui)), QWidget)

    def get_control_widget(self):
        u"""
        :return:
        """
        return wrapInstance(_ptr(omui.MQtUtil.findControl(self.ui)), QWidget)

    def get_layout_widget(self):
        u"""
        :return:
        """
        return wrapInstance(_ptr(omui.MQtUtil.findLayout(self.ui)), QWidget)

    def get_item_widget(self):
        u"""
        :return:
        """
        return wrapInstance(_ptr(omui.MQtUtil.findMenuItem(self.ui)), QWidget)

    def __melobject__(self):
        u"""
        获得mel对象

        :return:
        """
        return self.ui

    def __str__(self):
        return "UI({})".format(repr(self.ui))

    def __repr__(self):
        return "UI({})".format(repr(self.ui))

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


other_node_cls_table = dict()


def other_node_cls(api_type_str):
    def _node(cls):
        other_node_cls_table[api_type_str] = cls
        return cls

    return _node


_u_str_t = type('')
_any_str_t = (bytes, _u_str_t)


def new_object(o):
    if isinstance(o, BaseType):
        return o
    if isinstance(o, _any_str_t):
        try:
            ref = make_ref(o)
        except RuntimeError as ex:
            raise CPMelException("对象 {} 不存在".format(repr(o)))
        # 优先判断组件和属性减少判断次数
        # 优先判断组件因为组件的调用次数可能比较多
        if ref.ref_type == cr.TComponent:
            return Component(ref)
        if ref.ref_type == cr.TPlug:
            return Attr(ref)

        # 创建节点
        o = ref.unsafe_m_selection_list().getDependNode(0)
        other_cls = other_node_cls_table.get(o.apiTypeStr)
        if other_cls is None:
            if ref.ref_type == cr.TDependencyNode:
                return Node(ref)
            else:
                if o.hasFn(om2.MFn.kShape):
                    return Shape(ref)
                if o.hasFn(om2.MFn.kTransform):
                    return Transform(ref)
                return DagNode(ref)
        return other_cls(ref)
    raise TypeError


from cpmel._args_conv import arg_conv

import cpmel._object_types.other as _

del _
